/* $Workfile: T2p5x4MfR2.c $       $Revision: 13 $      $Modtime: 6/07/06 10:57a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2P5X4MFR2.C

    A simple example console application demonstrating how to use the OTX
    logical devices in OTXSPM2 to generate and detect MF R2 tones.
    
   The applications works with the following board combinations:

      Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-55x4-ASM,

    or OTX PCI Express series boards,

      Thor-2-PCI-Express + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Express + Vidar-55x4-ASM,

    or Thor-2-PCMCIA-Pro,

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines and expects them to be looped with an external
    loop cable.

    It also creates various tone device on the DSP, which are
    used to detect and generate tones.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1L          // Task Reference for a non-blocking call
#define DEMO_NO_OF_LIS 2
#define DEMO_NO_OF_LOGICAL_DEVS 64
#define DEMO_DEFAULT_LOG_FILE "T2p5x4MfR2.log"
#define DEMO_DEFAULT_BOARD_NO 0


typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  OTX_INT32       nBoardType;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;      // T1 or E1 mode
  OtxDspSpmE      eProgramModule; // Standard Program Module to be loaded in the DSP (from the Lib)
  int             boardNo;
  int             coDsp;
  OTX_BOOL        bUseDspDebugger;
  int             nAsm0Highway; 
  int             nCodecHighway;
#define DEMO_CODEC_HIGHWAY   14

} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoConnectCodecSpeakerSource(OTX_INT32 ts);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_INT32   g_nCodecTs = 0;       // Dsp Highway timeslot to connect the Codec to

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2];        // Only used for Thor-4-PCI-Plus and Thor-8-PCI(-Plus); not for Thor-2-PCI
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[2];
  OTX_HANDLE hCodec[4];
  OTX_HANDLE hDsp[4];
  OTX_HANDLE hTss;
} g_hPhysDevs;


struct {
  OTX_HANDLE hDspLi[64];
  OTX_HANDLE hLiDsp[64];
  OTX_HANDLE hTssCodec[4]; 
} g_hPipes;


struct {
  OTX_HANDLE hSingleToneDetector[64];
  OTX_HANDLE hMfR2BackwardDetector[64];
  OTX_HANDLE hMfR2ForwardDetector[64]; 
  OTX_HANDLE hSineWaveSender[64];
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DTMF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);						

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus)

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

  // Get handles to T1/E1 line interface devices #0 and #1
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < DEMO_NO_OF_LIS; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < DEMO_NO_OF_LIS; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      // On the Thor-8-PCI and Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < DEMO_NO_OF_LIS; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Get handles to Codec devices #0 - #3
    for (i = 0; i < 4; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      // Set correct companded law (u-law or a-law)
      result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec[i], g_demoEnv.law); DemoCheckRc(g_hPhysDevs.hCodec[i], result);
    }
  }


  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    OTX_UINT32   eDevType;
    OTX_ITERATOR itrChildTypes;
    OtxDrvIteratorBegin(&itrChildTypes);
    while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
      OTX_BOOL bOpenAsm = OTX_FALSE;
      switch (eDevType) {
      case OTX_DEVICE_VIDAR_5X4_ASM_V90:
        DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
        DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_55x4_ASM:
        DemoLogf("Opening Vidar-55x4-ASM board\n");
        bOpenAsm = OTX_TRUE;

        // Verify that the system supports the Vidar-55x4-ASM board
        {
          char szCpldBuildNumbers[256];
          char szBrdBuildNumber[16];
          OTX_INT32 nBoardBuildNo;
          
          if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
            
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
              DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
              DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 11) {
              DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
              DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
              DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 5) {
              DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          }
          
        }

        break;
      }
      if (bOpenAsm) {
        // Get a handle to to the ASM Daughter Board
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
    }
      
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
    } 
  } else {
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
    } 
  }

  // Load DSP Program
#if 1 // Load DSP Program from OtxDrv.lib
  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    assert(g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM2);
    DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", nDsp);
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);

  }
#endif
#if 0 // Load from file
  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    DemoLogf("Loading program file OtxSpm2.out to DSP#%d \n", nDsp);
    result = OtxDspRunProgram(g_hPhysDevs.hDsp[nDsp], "OtxSpm2.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  }
#endif

  if (g_demoEnv.bUseDspDebugger) {
    int c;
    DemoLogf("************************************************************\n\n");
    DemoLogf("Now please load and run DSP program to all of the DSPs\n");
    DemoLogf("in your DSP Debugger.\n\n");
    DemoLogf("After all DSPs are running:\n");
    DemoLogf("  Press 'r' to resynchronize the communication between the DSPs and the host\n");
    DemoLogf("  Press 's' to skip the resynchronization step\n");
    c = OsaConsoleGetCh();
    DemoLogf("\n");

    if (c == 'r' || c == 'R') {
      // At this point we need to tell the host to restart the programs 
      // (to issue the initial IO commands to the DSP again)
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        DemoLogf("Restarting program on DSP#%d\n", i);
        result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      }
      // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
    }
    
  }


}


/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < DEMO_NO_OF_LIS; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI-(Plus)

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

    // If board type is Thor-4-PCI-Plus then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  // If board type is Thor-8-PCI-Plus then enable Quad E1T1 device
  if ( g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS ) {

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }


  for (i = 0; i < DEMO_NO_OF_LIS; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;

  for (i = 0; i < DEMO_NO_OF_LOGICAL_DEVS; i++) {
    nDsp = i/32;
    
    // Create Sinewave Sender Logical Devices
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_TONE_SINEWAVE_GENERATOR, g_hEventQueue, i, &g_hLogicDevs.hSineWaveSender[i]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
    
    // Create MF Detector Logical Device for MF R2 Forward tones
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_TONE_MF_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hMfR2ForwardDetector[i]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
    
    // Create Dual Tone Detector Logical Device for MF R2 Backward tones
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_TONE_DUAL_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hMfR2BackwardDetector[i]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
    
    // Create Single Tone Detector Logical Device for MF R2 Supervisory tones
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_TONE_SINGLE_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hSingleToneDetector[i]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  }

}

/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 nLi;
  OTX_INT32 nDev;
  OTX_INT32 nDsp;
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < DEMO_NO_OF_LOGICAL_DEVS; i++) {

    result = OtxDrvCreatePipe(64, &g_hPipes.hLiDsp[i]); DemoCheckRc(g_hPipes.hLiDsp[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspLi[i]); DemoCheckRc(g_hPipes.hDspLi[i], result);

  }                  

  // Connect pipes for the SineWave Senders (MF) and MF Detectors
  for (nDev = 0; nDev < DEMO_NO_OF_LOGICAL_DEVS; nDev++) {
    //if (nDev%32 == 0) continue; // Skip TS0

    // Create pipes per Li. One pipe to each DSP
    // Li0 Ts1 <=> DSP0
    // Li1 Ts1 <=> DSP2
      ts = nDev%32;
      nDsp = nDev/32;
      nLi = nDev/32;

      // Connect Pipes
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hLiDsp[nDev], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiDsp[nDev], result);
      result = OtxDrvConnectPipe(g_hPipes.hDspLi[nDev], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspLi[nDev], result);

      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
        result = OtxDrvConnectPipe(g_hPipes.hLiDsp[nDev], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiDsp[nDev], result);
        result = OtxDrvConnectPipe(g_hPipes.hDspLi[nDev], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hDspLi[nDev], result);
      } else {
        result = OtxDrvConnectPipe(g_hPipes.hLiDsp[nDev], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_hPipes.hLiDsp[nDev], result);
        result = OtxDrvConnectPipe(g_hPipes.hDspLi[nDev], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_hPipes.hDspLi[nDev], result);
      }

  }


  for (i = 0; i < DEMO_NO_OF_LOGICAL_DEVS; i++) {

    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------

    // Create the following Connections in receive direction:
    //         +----> MfDetector     (MF R2 Forward tones)
    //  Li     +----> DualDetector   (MF R2 Backward tones)
    //         +----> SingleDetector (MF R2 Supervisory tone)
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDsp[i], -1, g_hLogicDevs.hMfR2ForwardDetector[i],   -1); DemoCheckRc(g_hPipes.hLiDsp[i], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDsp[i], -1, g_hLogicDevs.hMfR2BackwardDetector[i],   -1); DemoCheckRc(g_hPipes.hLiDsp[i], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDsp[i], -1, g_hLogicDevs.hSingleToneDetector[i],   -1); DemoCheckRc(g_hPipes.hLiDsp[i], result);

    // Make the following connections in transmit direction:
    // SinewaveSender   -----> Li
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hSineWaveSender[i], -1, g_hPipes.hDspLi[i], -1); DemoCheckRc(g_hLogicDevs.hSineWaveSender[i], result);

  }

  // Connect Codecs (speakers)
  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      // Connect the DSP source pipes to the to Codec, to allow listening to the generated tones
      // from the SinewaveGenerator or the DTMF generator.
      // DSP0 => Codec0
      // DSP1 => Codec1
      // DSP2 => Codec2
      // DSP3 => Codec3
      
      result = OtxDrvCreatePipe(64, &g_hPipes.hTssCodec[nDsp]); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
      
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[g_nCodecTs] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hTss, nDsp + g_demoEnv.nAsm0Highway, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);

      ts = 1+nDsp;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, g_demoEnv.nCodecHighway, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
    }  
    
  }


}


/*******************************************************************************

  @func
    Connect Codec speaker source.

*******************************************************************************/
void DemoConnectCodecSpeakerSource(
  OTX_INT32 ts
)
{
  OTX_INT32 nDsp;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];

  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    result = OtxDrvDisconnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hTssCodec[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hTss, nDsp + g_demoEnv.nAsm0Highway, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hTssCodec[nDsp], result);
    
  }        

}

/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;
  
  for (i = 0; i < DEMO_NO_OF_LOGICAL_DEVS; i++) {

    {
      OTX_UINT32 nThreshold;
      if (i == 0) {
        result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, &nThreshold); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
        DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
        
        result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY, &nThreshold); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
        DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY  = %d\n", i, nThreshold);
        
        result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, &nThreshold); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
        DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
      }
      
      // Modifing default thresholds
      nThreshold = 1000;
      //DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, nThreshold); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
      
      nThreshold = 200;
      //DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_VARIANCE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY, nThreshold); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
      
      nThreshold = 200;
      //DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, nThreshold); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
      
    }
    
    {
      
      OTX_UINT32    sqFreq[1];       // An array of up to 6 frequencies to be detected.
      OTX_INT32     coFreq = 1;       // Number of elements in <p sqFreq>
      
#if 0 // Include this code to get a more status messages from the Single Tone Detector
      result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_VERBOSE_STATUS, 5); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
#endif
      
      sqFreq[0] = 2560;  // Freq A in Hz
      
      
      // Initialize the Generic Single Tone Detector. Use a couple of DTMF tone frequencies as an example
      result =  OtxToneSingleDetectorInit(
        g_hLogicDevs.hSingleToneDetector[i],           // Handle to the Logical device
        g_demoEnv.law,    // Data Format of the Monitored Data Stream
        sqFreq,           // An array of up to 6 frequencies to be detected.
        coFreq,           // Number of elements in <p sqFreq>
        OTX_TRUE,         // If OTX_TRUE, generate events and store the received digits
        OTX_TRUE,         // If OTX_TRUE the detector generates an event when a tone ends
        250                // Minimum duration (in milliseconds) required to validate a dual tone
        );
      DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
      
    }


    
    {

      OTX_UINT32    sqFreq[6];       // An array of up to 6 frequencies to be detected.
      OTX_INT32     coFreq = 6;       // Number of elements in <p sqFreq>
      
      sqFreq[0] = 540;    // in Hz
      sqFreq[1] = 660;    // in Hz       
      sqFreq[2] = 780;    // in Hz
      sqFreq[3] = 900;    // in Hz  
      sqFreq[4] = 1020;   // in Hz
      sqFreq[5] = 1140;   // in Hz
      
      // Initialize the Generic Dual Tone Detector. Use a couple of DTMF tone frequencies as an example
      result =  OtxToneDualDetectorInit(
        g_hLogicDevs.hMfR2BackwardDetector[i],           // Handle to the Logical device
        g_demoEnv.law,    // Data Format of the Monitored Data Stream
        sqFreq,           // An array of up to 6 frequencies to be detected.
        coFreq,           // Number of elements in <p sqFreq>
        OTX_TRUE,         // If OTX_TRUE, generate events and store the received digits
        OTX_TRUE,         // If OTX_TRUE the detector generates an event when a tone ends
        12                // Minimum duration required to validate a digit
        );
      DemoCheckRc(g_hLogicDevs.hMfR2BackwardDetector[i], result);
    }
    
    
#if 0 // Include this code to get a more status messages from the Mf Detector
    if (i==0) {
      result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hMfR2ForwardDetector[i], OTX_ATTR_MF_DETECTOR_VERBOSE_STATUS, 1); DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
    }
#endif
    
    if (i==0) 
    {
      OTX_UINT32 nThreshold;
      
      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfR2ForwardDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TONE_ENERGY, &nThreshold); DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
      
      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfR2ForwardDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_PAUSE_ENERGY, &nThreshold); DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
      
      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfR2ForwardDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST, &nThreshold); DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST        = %d\n", i, nThreshold);
      
      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfR2ForwardDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_RELATIVE_3RD, &nThreshold); DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_RELATIVE_3RD = %d\n", i, nThreshold);
    }
    
#if 0 // To make the MF detector more stringent lower the TWIST threshold closer to 100
    {
      OTX_UINT32 nThreshold;
      
      nThreshold = 110;
      DemoLogf("MF Detector[%d]: Setting OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST        = %d\n", i, nThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hMfR2ForwardDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST, nThreshold); DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
    }
#endif
    
    
    // Initialize the MF Tone Detector 
    result = OtxToneMfDetectorInit(
      g_hLogicDevs.hMfR2ForwardDetector[i], 
      g_demoEnv.law,
      OTX_TONE_MF_R2_FORWARD,  // MF Tone type (OTX_TONE_MF_R1, OTX_TONE_MF_R2_FORWARD, OTX_TONE_MF_R2_BACKWARD)
      OTX_TRUE,           // If OTX_TRUE, generate events and store the received digits in the Event QueusFifo, If OTX_FALSE Don't store. 
      OTX_FALSE,          // If OTX_TRUE the detector generates an event when a tone ends    
      25);                // Minimum duration (in milliseconds) required to validate a digit 
    DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
    
  }
  
}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_NO_OF_LOGICAL_DEVS; i++) {
    if (i%32 == 0) continue; // Skip TS0
    result = OtxDrvEnable(g_hLogicDevs.hMfR2ForwardDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hMfR2ForwardDetector[i], result);
    result = OtxDrvEnable(g_hLogicDevs.hMfR2BackwardDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hMfR2BackwardDetector[i], result);
    result = OtxDrvEnable(g_hLogicDevs.hSingleToneDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i], result);
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*********************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Read a Driver Event
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  //DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_MF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("MfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("MfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        switch (pEventData->m_nParam) {
        case 0:  // 0  (941Hz/1336Hz)
        case 1:  // 1  (697Hz/1209Hz)
        case 2:  // 2  (697Hz/1336Hz)
        case 3:  // 3  (697Hz/1477Hz)
        case 4:  // 4  (770Hz/1209Hz)
        case 5:  // 5  (770Hz/1336Hz)
        case 6:  // 6  (770Hz/1477Hz)
        case 7:  // 7  (852Hz/1209Hz)
        case 8:  // 8  (852Hz/1336Hz)
        case 9:  // 9  (852Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 10: // A  (697Hz/1633Hz)
        case 11: // B  (770Hz/1633Hz)
        case 12: // C  (852Hz/1633Hz)
        case 13: // D  (941Hz/1633Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key %C \n", pEventData->m_nSourceId, pEventData->m_nParam - 10 + 'A');
          break;
        case 14: // *  (941Hz/1209Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '*'\n", pEventData->m_nSourceId);
          break;
        case 15: // #  (941Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '#'\n", pEventData->m_nSourceId);
          break;
        default: 
          DemoLogf("DtmfDetector[%d] Unrecognized DTMF tone (%d)\n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        }

      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DUAL_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dual Tone Detector (Backward Tone)
      if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_TONE) {
        OTX_INT32 nDigit = -1;
        switch (pEventData->m_nParam) {
        case 1: nDigit = 15; break;
        case 2: nDigit = 14; break;
        case 3: nDigit = 10; break;
        case 4: nDigit = 13; break;
        case 5: nDigit = 9; break;
        case 6: nDigit = 6; break;
        case 7: nDigit = 12; break;
        case 8: nDigit = 8; break;
        case 9: nDigit = 5; break;
        case 10: nDigit = 3; break;
        case 11: nDigit = 11; break;
        case 12: nDigit = 7; break;
        case 13: nDigit = 4; break;
        case 14: nDigit = 2; break;
        case 15: nDigit = 1; break;
        default:
          break;
        }
        if (nDigit == -1) {
          DemoLogf("DualDetector[Li%d:TS%d] unknown code %d \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nParam);
        } else {
          DemoLogf("DualDetector[Li%d:TS%d] detected MFR2 Backward digit %d \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, nDigit);
        }
      } else if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_PAUSE) {
        DemoLogf("DualDetector[Li%d:TS%d] detected Pause\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32);
      } else if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_VERBOSE_STATUS) {
        DemoLogf("DualDetector[Li%d:TS%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        // Unexpected event from this device
        DemoLogf("DualDetector[Li%d:TS%d] Unexpected Event: %s \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, msgStr);
      }
    } else {
      DemoLogf("DualDetector[Li%d:TS%d] Invalid Event: %ld \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_MF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_DIGIT) {
        DemoLogf("MfDetector[Li%d:TS%d] detected MFR2 Forward digit %.2d \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nParam);
      } else if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_VERBOSE_STATUS){
        DemoLogf("MfDetector[Li%d:TS%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        if (pEventData->m_nParam) {
          DemoLogf("MfDetector[Li%d:TS%d] Event: %s (%d)\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("MfDetector[Li%d:TS%d] Event: %s\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, msgStr);
        }
      }
    } else {
      DemoLogf("MfDetector[Li%d:TS%d] Invalid Event: %ld \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_SINGLE_DETECTOR:

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Single Tone Detector
      if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_TONE) {
        // sqFreq[0] = 2560;  // Freq A in Hz

        switch (pEventData->m_nParam) {
        case 0:
          DemoLogf("SingleDetector[Li%d:TS%d] detected MF R2 Supervisory Tone\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32);
          break;
        default:
          DemoLogf("SingleDetector[Li%d:TS%d] detected unrecognized tone (%d) \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nParam);
          break;
        }

      } else if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_PAUSE) {
        DemoLogf("SingleDetector[Li%d:TS%d] detected Pause\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32);
      } else if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_VERBOSE_STATUS) {
        DemoLogf("SingleDetector[Li%d:TS%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        // Unexpected event from this device
        DemoLogf("SingleDetector[Li%d:TS%d] Unexpected Event: %s \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, msgStr);
      }
    } else {
      DemoLogf("SingleDetector[Li%d:TS%d] Invalid Event: %ld \n", pEventData->m_nSourceId/32, pEventData->m_nSourceId%32, pEventData->m_nCode);
    }
    break;


  
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a DTMF tone).

*******************************************************************************/
void DemoSendSineWaveDtmfTone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nDtmfDigit = 0;

  OtxToneSineWaveDataS sqSineWaveData[5];
  
  sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[0].nPhase       = 0;  
  
  sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[1].nPhase       = 0;  
  
  switch (nDtmfDigit) {
  case 0:
    // DTMF digit 0 
    sqSineWaveData[0].nFrequency   = 941; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 1:
    // DTMF digit 1 
    sqSineWaveData[0].nFrequency   = 697; // in Hz
    sqSineWaveData[1].nFrequency   = 1209; // in Hz       
    break;
  case 2:
    // DTMF digit 2 
    sqSineWaveData[0].nFrequency   = 697; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 3:
    // DTMF digit 3 
    sqSineWaveData[0].nFrequency   = 697; // in Hz
    sqSineWaveData[1].nFrequency   = 1477; // in Hz       
    break;
  case 4:
    // DTMF digit 4 
    sqSineWaveData[0].nFrequency   = 770; // in Hz
    sqSineWaveData[1].nFrequency   = 1209; // in Hz       
    break;
  case 5:
    // DTMF digit 5 
    sqSineWaveData[0].nFrequency   = 770; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 6:
    // DTMF digit 6
    sqSineWaveData[0].nFrequency   = 770; // in Hz
    sqSineWaveData[1].nFrequency   = 1477; // in Hz       
    break;
  case 7:
    // DTMF digit 7 
    sqSineWaveData[0].nFrequency   = 852; // in Hz
    sqSineWaveData[1].nFrequency   = 1209; // in Hz       
    break;
  case 8:
    // DTMF digit 8 
    sqSineWaveData[0].nFrequency   = 852; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 9:
    // DTMF digit 9
    sqSineWaveData[0].nFrequency   = 852; // in Hz
    sqSineWaveData[1].nFrequency   = 1477; // in Hz       
    break;
  }
  
  DemoLogf("Sending DTMF tone %d (%d Hz / %d Hz)\n", nDtmfDigit, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
  result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
    g_demoEnv.law, 
    sqSineWaveData,
    2, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  
  //DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  OsaThreadSleep(500);
  
  //DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  if (++nDtmfDigit > 9) {
    nDtmfDigit = 0;
  }
  
    

}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a MF T1 R1).

*******************************************************************************/
void DemoSendSineWaveMfR1T1Tone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nMfDigit = 0;
  char *szMfDigitStr = "?";
  OtxToneSineWaveDataS sqSineWaveData[5];
  
  sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[0].nPhase       = 0;  
  
  sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[1].nPhase       = 0;  
  
  switch (nMfDigit) {
  case 0:
    // MF digit 0 
    szMfDigitStr = "0";
    sqSineWaveData[0].nFrequency   = 1300; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 1:
    // MF digit 1 
    szMfDigitStr = "1";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 900; // in Hz       
    break;
  case 2:
    // MF digit 2 
    szMfDigitStr = "2";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1100; // in Hz       
    break;
  case 3:
    // MF digit 3 
    szMfDigitStr = "3";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1100; // in Hz       
    break;
  case 4:
    // MF digit 4 
    szMfDigitStr = "4";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1300; // in Hz       
    break;
  case 5:
    // MF digit 5 
    szMfDigitStr = "5";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1300; // in Hz       
    break;
  case 6:
    // MF digit 6
    szMfDigitStr = "6";
    sqSineWaveData[0].nFrequency   = 1100; // in Hz
    sqSineWaveData[1].nFrequency   = 1300; // in Hz       
    break;
  case 7:
    // MF digit 7 
    szMfDigitStr = "7";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 8:
    // MF digit 8 
    szMfDigitStr = "8";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 9:
    // MF digit 9
    szMfDigitStr = "9";
    sqSineWaveData[0].nFrequency   = 1100; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 10:
    // MF KP
    szMfDigitStr = "KP(10)";
    sqSineWaveData[0].nFrequency   = 1100; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 11:
    // MF ST
    szMfDigitStr = "ST(11)";
    sqSineWaveData[0].nFrequency   = 1500; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 12:
    // MF ST1(x)
    szMfDigitStr = "ST1(12)";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 13:
    // MF ST2(y)
    szMfDigitStr = "ST2(13)";
    sqSineWaveData[0].nFrequency   = 1300; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  case 14:
    // MF ST3(z)
    szMfDigitStr = "ST3(14)";
    sqSineWaveData[0].nFrequency   = 700; // in Hz
    sqSineWaveData[1].nFrequency   = 1700; // in Hz       
    break;
  }
  
  DemoLogf("Sending R1 MF tone %s (%d Hz / %d Hz)\n", szMfDigitStr, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
  result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
    g_demoEnv.law, 
    sqSineWaveData,
    2, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  
  //DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  OsaThreadSleep(500);
  
  //DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  if (++nMfDigit > 14) {
    nMfDigit = 0;
  }
  
    

}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a MF E1 R2 Forward).

*******************************************************************************/
void DemoSendSineWaveMfR2E1ForwardTone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nMfDigit = 1;
  char *szMfDigitStr = "?";
  OtxToneSineWaveDataS sqSineWaveData[5];
  
  sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[0].nPhase       = 0;  
  
  sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[1].nPhase       = 0;  
  
  switch (nMfDigit) {
  case 1:
    // MF digit 1 
    szMfDigitStr = "1";
    sqSineWaveData[0].nFrequency   = 1380; // in Hz
    sqSineWaveData[1].nFrequency   = 1500; // in Hz       
    break;
  case 2:
    // MF digit 2 
    szMfDigitStr = "2";
    sqSineWaveData[0].nFrequency   = 1380; // in Hz
    sqSineWaveData[1].nFrequency   = 1620; // in Hz       
    break;
  case 3:
    // MF digit 3 
    szMfDigitStr = "3";
    sqSineWaveData[0].nFrequency   = 1500; // in Hz
    sqSineWaveData[1].nFrequency   = 1620; // in Hz       
    break;
  case 4:
    // MF digit 4 
    szMfDigitStr = "4";
    sqSineWaveData[0].nFrequency   = 1380; // in Hz
    sqSineWaveData[1].nFrequency   = 1740; // in Hz       
    break;
  case 5:
    // MF digit 5 
    szMfDigitStr = "5";
    sqSineWaveData[0].nFrequency   = 1500; // in Hz
    sqSineWaveData[1].nFrequency   = 1740; // in Hz       
    break;
  case 6:
    // MF digit 6
    szMfDigitStr = "6";
    sqSineWaveData[0].nFrequency   = 1620; // in Hz
    sqSineWaveData[1].nFrequency   = 1740; // in Hz       
    break;
  case 7:
    // MF digit 7 
    szMfDigitStr = "7";
    sqSineWaveData[0].nFrequency   = 1380; // in Hz
    sqSineWaveData[1].nFrequency   = 1860; // in Hz       
    break;
  case 8:
    // MF digit 8 
    szMfDigitStr = "8";
    sqSineWaveData[0].nFrequency   = 1500; // in Hz
    sqSineWaveData[1].nFrequency   = 1860; // in Hz       
    break;
  case 9:
    // MF digit 9
    szMfDigitStr = "9";
    sqSineWaveData[0].nFrequency   = 1620; // in Hz
    sqSineWaveData[1].nFrequency   = 1860; // in Hz       
    break;
  case 10:
    // MF digit a
    szMfDigitStr = "10";
    sqSineWaveData[0].nFrequency   = 1740; // in Hz
    sqSineWaveData[1].nFrequency   = 1860; // in Hz       
    break;
  case 11:
    // MF digit b
    szMfDigitStr = "11";
    sqSineWaveData[0].nFrequency   = 1380; // in Hz
    sqSineWaveData[1].nFrequency   = 1980; // in Hz       
    break;
  case 12:
    // MF digit c
    szMfDigitStr = "12";
    sqSineWaveData[0].nFrequency   = 1500; // in Hz
    sqSineWaveData[1].nFrequency   = 1980; // in Hz       
    break;
  case 13:
    // MF digit d
    szMfDigitStr = "13";
    sqSineWaveData[0].nFrequency   = 1620; // in Hz
    sqSineWaveData[1].nFrequency   = 1980; // in Hz       
    break;
  case 14:
    // MF digit e
    szMfDigitStr = "14";
    sqSineWaveData[0].nFrequency   = 1740; // in Hz
    sqSineWaveData[1].nFrequency   = 1980; // in Hz       
    break;
  case 15:
    // MF digit f
    szMfDigitStr = "15";
    sqSineWaveData[0].nFrequency   = 1860; // in Hz
    sqSineWaveData[1].nFrequency   = 1980; // in Hz       
    break;
  }
  
  DemoLogf("Sending R2 MF Forward tone %s (%d Hz / %d Hz)\n", szMfDigitStr, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
  result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
    g_demoEnv.law, 
    sqSineWaveData,
    2, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  
  //DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  OsaThreadSleep(250);
  //OtxDrvDbgDump(g_hPhysDevs.hDsp[1], 0);
  OsaThreadSleep(250);
  
  //DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  if (++nMfDigit > 15) {
    nMfDigit = 1;
  }
  
  OsaThreadSleep(100);
  //OtxDrvDbgDump(g_hPhysDevs.hDsp[1], 0);
    
}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a MF E1 R2 Backward).

*******************************************************************************/
void DemoSendSineWaveMfR2E1BackwardTone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nMfDigit = 1;
  char *szMfDigitStr = "?";
  OtxToneSineWaveDataS sqSineWaveData[5];
  
  sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[0].nPhase       = 0;  
  
  sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[1].nPhase       = 0;  
  
  switch (nMfDigit) {
  case 1:
    // MF digit 1 
    szMfDigitStr = "1";
    sqSineWaveData[0].nFrequency   = 1140; // in Hz
    sqSineWaveData[1].nFrequency   = 1020; // in Hz       
    break;
  case 2:
    // MF digit 2 
    szMfDigitStr = "2";
    sqSineWaveData[0].nFrequency   = 1140; // in Hz
    sqSineWaveData[1].nFrequency   = 900; // in Hz       
    break;
  case 3:
    // MF digit 3 
    szMfDigitStr = "3";
    sqSineWaveData[0].nFrequency   = 1020; // in Hz
    sqSineWaveData[1].nFrequency   = 900; // in Hz       
    break;
  case 4:
    // MF digit 4 
    szMfDigitStr = "4";
    sqSineWaveData[0].nFrequency   = 1140; // in Hz
    sqSineWaveData[1].nFrequency   = 780; // in Hz       
    break;
  case 5:
    // MF digit 5 
    szMfDigitStr = "5";
    sqSineWaveData[0].nFrequency   = 1020; // in Hz
    sqSineWaveData[1].nFrequency   = 780; // in Hz       
    break;
  case 6:
    // MF digit 6
    szMfDigitStr = "6";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 780; // in Hz       
    break;
  case 7:
    // MF digit 7 
    szMfDigitStr = "7";
    sqSineWaveData[0].nFrequency   = 1140; // in Hz
    sqSineWaveData[1].nFrequency   = 660; // in Hz       
    break;
  case 8:
    // MF digit 8 
    szMfDigitStr = "8";
    sqSineWaveData[0].nFrequency   = 1020; // in Hz
    sqSineWaveData[1].nFrequency   = 660; // in Hz       
    break;
  case 9:
    // MF digit 9
    szMfDigitStr = "9";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 660; // in Hz       
    break;
  case 10:
    // MF digit a
    szMfDigitStr = "10";
    sqSineWaveData[0].nFrequency   = 780; // in Hz
    sqSineWaveData[1].nFrequency   = 660; // in Hz       
    break;
  case 11:
    // MF digit b
    szMfDigitStr = "11";
    sqSineWaveData[0].nFrequency   = 1140; // in Hz
    sqSineWaveData[1].nFrequency   = 540; // in Hz       
    break;
  case 12:
    // MF digit c
    szMfDigitStr = "12";
    sqSineWaveData[0].nFrequency   = 1020; // in Hz
    sqSineWaveData[1].nFrequency   = 540; // in Hz       
    break;
  case 13:
    // MF digit d
    szMfDigitStr = "13";
    sqSineWaveData[0].nFrequency   = 900; // in Hz
    sqSineWaveData[1].nFrequency   = 540; // in Hz       
    break;
  case 14:
    // MF digit e
    szMfDigitStr = "14";
    sqSineWaveData[0].nFrequency   = 780; // in Hz
    sqSineWaveData[1].nFrequency   = 540; // in Hz       
    break;
  case 15:
    // MF digit f
    szMfDigitStr = "15";
    sqSineWaveData[0].nFrequency   = 660; // in Hz
    sqSineWaveData[1].nFrequency   = 540; // in Hz       
    break;
  }
  
  DemoLogf("Sending R2 MF Backward tone %s (%d Hz / %d Hz)\n", szMfDigitStr, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
  result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
    g_demoEnv.law, 
    sqSineWaveData,
    2, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  
  //DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  OsaThreadSleep(500);
  
  //DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  if (++nMfDigit > 15) {
    nMfDigit = 1;
  }
  
  
}


/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a generic tone. Prompts for 
  frequencies, attenuation and duration.

*******************************************************************************/
void DemoSendGenericSineWave(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  OTX_INT32 i;
  static OTX_INT32 nSineWaveComponents = 1;
  static OTX_INT32 nFrequency = 2560;
  static OTX_INT32 nAttenuation = 0;  // <=0
  static OTX_INT32 nDuration = 1000;  // in milliseconds
  char strBuf[64];
  OtxToneSineWaveDataS sqSineWaveData[5];
  
  DemoLogf("Please enter duration generated tone (in milliseconds) [%d]", nDuration);
  OsaConsoleGetStr(strBuf);
  if (*strBuf) {
    nDuration = strtol(strBuf, NULL, 0);
  }
  
  if (nDuration < 1) {
    DemoLogf("Invalid range. Operation aborted.\n");
    return;
  }
  
  DemoLogf("Please enter number of frequencies in the generated tone (1-5) [%d]", nSineWaveComponents);
  OsaConsoleGetStr(strBuf);
  if (*strBuf) {
    nSineWaveComponents = strtol(strBuf, NULL, 0);
  }
  
  if ((nSineWaveComponents < 1) || (nSineWaveComponents > 5)) {
    DemoLogf("Invalid range. Operation aborted.\n");
    return;
  }
  
  
  for (i = 0; i < nSineWaveComponents; i++) {
    
    DemoLogf("[%d]: Frequency? (1-8000) [%d]", i, nFrequency);
    OsaConsoleGetStr(strBuf);
    if (*strBuf) {
      nFrequency = strtol(strBuf, NULL, 0);
    }
    if ((nFrequency < 1) || (nFrequency > 8000)) {
      DemoLogf("Invalid range. Operation aborted.\n");
      return;
    }
    
    DemoLogf("[%d]: Attenuation? (<=0) [%+d]", i, nAttenuation);
    OsaConsoleGetStr(strBuf);
    if (*strBuf) {
      nAttenuation = strtol(strBuf, NULL, 0);
    }
    if (nAttenuation > 0) {
      DemoLogf("Invalid range. Operation aborted.\n");
      return;
    }
    
    sqSineWaveData[i].nFrequency   = nFrequency; // in Hz
    sqSineWaveData[i].nAttenuation = nAttenuation; 
    sqSineWaveData[i].nPhase       = 0;  
  }
  
  DemoLogf("Sending tone %d", (OTX_INT32)sqSineWaveData[0].nFrequency);
  for (i = 0; i < nSineWaveComponents; i++) {
    DemoLogf("/ %d ", (OTX_INT32)sqSineWaveData[i].nFrequency);
  }
  DemoLogf(" Hz. Duration = %d ms\n", nDuration);
  
  DemoLogf("Press <Enter> to start tone generation...\n");
  OsaConsoleGetStr(strBuf);
  
  result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
    g_demoEnv.law, 
    sqSineWaveData,
    nSineWaveComponents, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  
  //DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
  OsaThreadSleep(nDuration);
  
  //DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hSineWaveSender[nDialler], result);
  
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr      = NULL;
  envP->bExitOnError     = OTX_FALSE;
  envP->law              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->eT1E1Mode        = OTX_T1E1_MODE_T1;
  envP->eProgramModule   = OTX_DSP_SPM_OTXSPM2;
  envP->boardNo          = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType       = OTX_DEVICE_THOR_PCMCIA;
  envP->coDsp            = 4;
  envP->bUseDspDebugger  = OTX_FALSE;
  envP->nAsm0Highway     = -1; // N/A
  envP->nCodecHighway    = -1; // N/A


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'X') {   // DSP Debugger
          envP->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coDsp      = 2;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType = OTX_DEVICE_THOR_PCMCIA;
            envP->coDsp      = 2;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coDsp      = 2;
            envP->nAsm0Highway  = 8;
            envP->nCodecHighway = 0;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coDsp      = 2;
            envP->nAsm0Highway  = 8;
            envP->nCodecHighway = 7;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coDsp      = 2;
            envP->nAsm0Highway  = 8;
            envP->nCodecHighway = 7;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}   


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[T|E|4|M|P]    : Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-PLUS with Vidar-5x4-ASM-PRO Board\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express with Vidar-5x4-ASM-PRO Board\n");
  DemoLogf("                  4 = Use Thor-4-PCI-PLUS with Vidar-5x4-ASM-PRO Board\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA-PRO Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-PLUS with Vidar-5x4-ASM-PRO Board\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /X           : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentLi = 0;
  static OTX_INT32      nCurrentDialer = 1;
  static OTX_INT32      nCurrentTs = 1;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" i   Send a generic tone from current dialer\n");
    DemoLogf(" j   Send one DTMF tone from current dialer\n");
    DemoLogf(" k   Send one R1 MF tone from current\n");
    DemoLogf(" l   Send one R2 MF Forward tone from current dialer\n");
    DemoLogf(" m   Send one R2 MF Backward tone from current dialer\n");
    DemoLogf(" n   Increment current dialer (dialer#%d (Li%d:TS%d) Selected\n", nCurrentDialer, nCurrentLi, nCurrentTs);
    DemoLogf(" N   Decrement current dialer (dialer#%d (Li%d:TS%d) Selected\n", nCurrentDialer, nCurrentLi, nCurrentTs);
    if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
      DemoLogf(" >   Increment TS for codecs (current TS=%d)\n", g_nCodecTs);
      DemoLogf(" <   Decrement TS for codecs (current TS=%d)\n", g_nCodecTs);
    }
    break;

  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;

  case ':':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < 1; nDsp++) {
        DemoLogf("Debug Dump for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
      }
    }
    break;

  case 'i':
    {
      DemoLogf("Generate a generic tone on Li%d:TS%d \n", nCurrentLi, nCurrentTs);

      DemoSendGenericSineWave(nCurrentDialer);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'j':
    {
      DemoLogf("Sending SineWaveTone on Li%d:TS%d \n", nCurrentLi, nCurrentTs);

      DemoSendSineWaveDtmfTone(nCurrentDialer);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'k':
    {
      DemoLogf("Sending SineWaveTone on Li%d:TS%d \n", nCurrentLi, nCurrentTs);

      DemoSendSineWaveMfR1T1Tone(nCurrentDialer);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'l':
    {
      DemoLogf("Sending SineWaveTone on Li%d:TS%d \n", nCurrentLi, nCurrentTs);

      DemoSendSineWaveMfR2E1ForwardTone(nCurrentDialer);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'm':
    {
      DemoLogf("Sending SineWaveTone on Li%d:TS%d \n", nCurrentLi, nCurrentTs);

      DemoSendSineWaveMfR2E1BackwardTone(nCurrentDialer);

      DemoLogf("Done.\n");
    }
    break;

  
  case 'z':
    {
      OtxT1E1TermModeE eTermMode;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    }
    break;


  case '0':
    // Select Li 0 as current
    nCurrentLi = 0;
    if (nCurrentDialer > 32) nCurrentDialer -= 32;
    nCurrentTs = nCurrentDialer%32;
    DemoLogf("Dialer#%d (Li%d, TS%d) Selected \n", nCurrentDialer, nCurrentLi, nCurrentTs);
    break;

  case '1':
    // Select Li 1 as current
    nCurrentLi = 1;
    if (nCurrentDialer < 32) nCurrentDialer += 32;
    nCurrentTs = nCurrentDialer%32;
    DemoLogf("Dialer#%d (Li%d, TS%d) Selected \n", nCurrentDialer, nCurrentLi, nCurrentTs);
    break;

  case '>':
    {
      if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
        g_nCodecTs++;
        if (g_nCodecTs > 31) {
          g_nCodecTs = 0;
        }
        
        DemoConnectCodecSpeakerSource(g_nCodecTs);
          
        DemoLogf("All codecs are now connected to TS%d (DSP highways)\n", g_nCodecTs);
      } else {
        DemoLogf("Function not supported by this board type\n");
      }
    }
    break;

  case '<':
    {
      if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
        g_nCodecTs--;
        if (g_nCodecTs < 0) {
          g_nCodecTs = 31;
        }
        
        DemoConnectCodecSpeakerSource(g_nCodecTs);
        
        DemoLogf("All codecs are now connected to TS%d (DSP highways)\n", g_nCodecTs);
      } else {
        DemoLogf("Function not supported by this board type\n");
      }
    }
    break;

  case 'n':
    {
      // Select current nCurrentDialer
      nCurrentDialer++;
      if (nCurrentDialer > 63) {
        // Wrap (and skip TS 0)
        nCurrentDialer = 1;
      } else if (nCurrentDialer%32 == 0) {
        // Li1:TS0 is signalling channel. Skip one up
        nCurrentDialer++;
      }
      
      nCurrentLi = nCurrentDialer/32;
      nCurrentTs = nCurrentDialer%32;
      DemoLogf("Dialer#%d (Li%d, TS%d) Selected \n", nCurrentDialer, nCurrentLi, nCurrentTs);

    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentDialer
      nCurrentDialer--;
      if (nCurrentDialer < 1) {
        // Wrap (and skip TS 0)
        nCurrentDialer = 63;
      } else if (nCurrentDialer%32 == 0) {
        // Li1:TS0 is signalling channel. Skip one down
        nCurrentDialer--;
      }
      
      nCurrentLi = nCurrentDialer/32;
      nCurrentTs = nCurrentDialer%32;
      DemoLogf("Dialer#%d (Li%d, TS%d) Selected \n", nCurrentDialer, nCurrentLi, nCurrentTs);
    }
    break;
    
        


  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("T2p5x4MfR2, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("MF R2 example program for OTX boards.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set Li0 as the clock source
  if (g_hPhysDevs.hVidarAsmController0) {
    // Thor-2-PCI, Thor-8-PCI
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
#if 1  // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif
  } else {
    // Thor-2-PCMCIA
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

#if 0 // For testing
  {
    OTX_INT32 i;
    
    for (i = 0; i < 3000; i++) {
      // Create and close the Logical devices needed by this application
      DemoLogf("Creating and closing logical devices [%d]...\n", i);
      DemoCreateLogicalDevices();
      DemoCloseLogicalDevices();
    }
  }
#endif

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // To let DemoIdleFunction start up
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", "/P2", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
