/* $Workfile: T2mRawDataSync.c $       $Revision: 31 $      $Modtime: 6/07/06 9:48a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2MRAWDATASYNC.C

    A simple example console application demonstrating sending two raw data
    files (one from each Line Interface) in a synchronized fashion (i.e. the 
    first byte of each file is transmitted in the same T1/E1 multiframe.

    The application also receives data from two timeslots (one from each
    Line Interface) and places the received data into raw data files.

    The program uses Raw Data Sender and the Raw Data Receiver with the 
    OTX Driver using the boards:
    
    The applications works with the following board combinations:

      Thor-2-PCI / Thor-8-PCI + Vidar-5x4-ASM,

    or OTX PCI Plus series boards,

      Thor-2-PCI-Plus / Thor-2-PCI-Express / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-CST,
      Thor-2-PCI-Plus / Thor-2-PCI-Express / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Plus / Thor-2-PCI-Express / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-EX,
      Thor-2-PCI-Plus / Thor-2-PCI-Express / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-55x4-ASM,

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or Sleipnir-1-PCI

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures up to 2 E1 Lines and expects them to be looped with an external
    loop cable.

    The Raw Data Receiver and the Raw Data Sender is implemented in the
    DSP device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#define DEMO_DEFAULT_LOG_FILE "T2mRawDataSync.log"

#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_DEFAULT_LI0_TX_FILE_NAME "Li0Tx.dat"
#define DEMO_DEFAULT_LI1_TX_FILE_NAME "Li1Tx.dat"
#define DEMO_DEFAULT_LI0_RX_FILE_NAME "Li0Rx.dat"
#define DEMO_DEFAULT_LI1_RX_FILE_NAME "Li1Rx.dat"

#define DEMO_DEFAULT_SETUP_TIME 250  // 250 milliseconds
#define DEMO_DEFAULT_HOLD_TIME 500

#define DEMO_RAW_TX_BUF_SIZE 1024

#define DEMO_TEST_STRING "abcdefg"

#define DEMO_DSP_MAX 2 


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL              bExitOnError;
  char                  cwd[OTX_MAX_PATH];
  char                  logFileName[OTX_MAX_PATH];
  char                  *pszLi0TxFileName;
  char                  *pszLi1TxFileName;
  char                  *pszLi0RxFileName;
  char                  *pszLi1RxFileName;
  OtxPhysicalDeviceTypeE nBoardType;
  int                    boardNo;
  int                    coLi;
  int                    coDsp;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_INT32              nLi0TxTs;
  OTX_INT32              nLi1TxTs;
  OTX_INT32              nLi0RxTs;
  OTX_INT32              nLi1RxTs;
  OtxDspSpmE             eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
  OTX_INT32              nRawDataCapacity;
  OTX_BOOL               bInteractiveMode;
  OTX_UINT8              nIdleCode;
  OTX_INT32              nLogVerbosity;
  OTX_INT32              nHoldTime;
  OTX_INT32              nSetupTime;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OTX_BOOL               bReceiveEqualizer; 
  OtxBrdClockSourceE     eClkSource;
  OTX_BOOL               bConfigure;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf1(char *formatStr, ...);
static void DemoLogf2(char *formatStr, ...);
static void DemoLogf3(char *formatStr, ...);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoExit(void);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;
OTX_BOOL    g_bExitFlag = OTX_FALSE;
OTX_UINT32  g_ThreadCount = 0;

OsaMutexS g_demoLogfCritSec;


struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Only used for Thor-4-PCI-Plus and Thor-8-PCI(-Plus)
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[2];
  OTX_HANDLE hDsp[DEMO_DSP_MAX];  // In the case of a Sleipnir board this is a handle to the burst device
} g_hPhysDevs;



struct {
  // External Pipes (The T1/E1 span) 
  OTX_HANDLE hLiIn;
  OTX_HANDLE hLiOut;

  // Pipes for Raw Data Sender/Receiver in the DSP
  OTX_HANDLE hDspLiRawData;
  OTX_HANDLE hLiDspRawData;
} g_hPipes[2];


struct sRawDataSender {
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_UINT32 nSourceId;

  //DWORD nThreadId;
  //HANDLE hThread;

  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_RAW_TX_BUF_SIZE];
  FILE       *pTxFile;
  OTX_INT32   cbTxBuf;
  OTX_INT32   coUnderRunErrors;
};

struct sRawDataReceiver {
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_UINT32 nSourceId;

  //DWORD nThreadId;
  //HANDLE hThread;

  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE       *pRxFile;
};

struct {
  struct sRawDataSender mRawDataSender[2];
  struct sRawDataReceiver mRawDataReceiver[2];
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-2 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.


  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI and Thor-8-PCI-Plus

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_PCI:
      DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening Thor-2-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      DemoLogf("Opening Thor-2-PCI-Express board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening Thor-4-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI:
      DemoLogf("Opening Thor-8-PCI board\n");
      break;
    case OTX_DEVICE_THOR_PCMCIA:
      DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      DemoLogf("Opening Thor-8-PCI-Plus board\n");
      break;
    case OTX_DEVICE_SLEIPNIR_1_PCI:
      DemoLogf("Opening Sleipnir-1-PCI-Plus board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    
  // set the board NOT to reset when we open sub devices - we will assume that the Li has already been setup.
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_RESET_ON_OPEN, g_demoEnv.bConfigure); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Get handles to T1/E1 line interface devices #0 and #1
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On the board the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI:
    {
      // On the Thor-8-PCI and Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }


  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    // Get handles to the physical DSP devices (on the PCMCIA board these devices are hosted by the PCMCIA board itself)
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI || g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_EXPRESS || OTX_DEVICE_THOR_4_PCI_PLUS || OTX_DEVICE_THOR_8_PCI || OTX_DEVICE_THOR_8_PCI_PLUS ) {
    // Get a handle to to the ASM Daughter Board
    OTX_UINT32   eDevType;
    OTX_ITERATOR itrChildTypes;
    OtxDrvIteratorBegin(&itrChildTypes);
    while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
      OTX_BOOL bOpenAsm = OTX_FALSE;
      switch (eDevType) {
      case OTX_DEVICE_VIDAR_5x_ASM:
        DemoLogf("Opening Vidar-5x4-ASM board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_CST:
        DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_V90:
        DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_EX:
        DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
        DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
        bOpenAsm = OTX_TRUE;
        break;
      case OTX_DEVICE_VIDAR_55x4_ASM:
        DemoLogf("Opening Vidar-55x4-ASM board\n");
        bOpenAsm = OTX_TRUE;

        // Verify that the system supports the Vidar-55x4-ASM board
        {
          char szCpldBuildNumbers[256];
          char szBrdBuildNumber[16];
          OTX_INT32 nBoardBuildNo;
          
          if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
            
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
              DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
              DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 11) {
              DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
            //  Verify firmware revision
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
              DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
              DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
              DemoLogf("       web site (www.odints.com)\n");
              DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
              DemoExitQuestion();
            }
            
            // Verify board build number
            result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
            nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
            DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
            if (nBoardBuildNo < 5) {
              DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
              DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
              DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
              DemoExitQuestion();
            }
            
          }
          
        }
        break;
      }
      if (bOpenAsm) {
        // Get a handle to to the ASM Daughter Board
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
    }
    
    // Get handles to the physical DSP devices (on the PCI board these devices are hosted by the ASM board)
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    printf("hello");
    // Get a handle to the physical Burst device.
    // To keep the changes to a minimum we use the hDsp variable as the handle
    // to the burst device. The burst device can, in this demo program, do
    // everything that the DSP device can do, except for loading a DSP program.
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_BURST,   i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]);   DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    
  } else {
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
  }

  // Disable DSP Serial port clocks (so that we can enable them synchronously after the DSP programs are loaded)
  result = OtxDrvSetAttributeValueBool(
    g_hPhysDevs.hThorController0, 
    OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
    OTX_TRUE
    ); 

  
  // Load DSP Program
  if (g_demoEnv.nBoardType != OTX_DEVICE_SLEIPNIR_1_PCI) {
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      
      switch (g_demoEnv.eProgramModule) {
      case OTX_DSP_SPM_OTXSPM1:
        DemoLogf2("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM1\n", i);
        break;
      case OTX_DSP_SPM_OTXHDLC:
        DemoLogf2("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC\n", i);
        break;
      default:
        DemoLogf2("Unsupport DSP program module (%d) for this demo program. Exiting.\n", g_demoEnv.eProgramModule);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
      }
      
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      
    }
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  if (g_demoEnv.bConfigure) {

    // Set default configuration data
    DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

    for (i = 0; i < g_demoEnv.coLi; i++) {

      // set the idle character for the line
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_TX_IDLE_CHARACTER, g_demoEnv.nIdleCode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_USE_TX_IDLE_CHARACTER, OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);

      // Set termination for the T1/E1 Front End
      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

      // Configure Transceivers
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      } else {
        // E1 mode
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }

      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
  }
}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI-(Plus)

  // Enable the board (Physical Device)
  DemoLogf2("Enable hThorController0\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // If board type is Thor-4-PCI-Plus then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS)) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices
    DemoLogf2("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    DemoLogf2("Enable g_hPhysDevs.hDsp[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}


void DemoCreateLogicalDevices(
/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

    Logical Devices can be classified into Control Device and Data Devices:

    - Control Device:  A logical device performing a control function. E.g. Event Queue.

    - Data Filter Device: A device implemented in Hardware, Firmware of Software
    which converts, processes, or manipulates serial data. Examples of data devices include
    Time-Space Switch, HDLC Controller, Tone Detector, Tone Generator, etc.

    Data Filter devices can be categorized into 3 basic types:

     - Data Source: Originates a Pipe or adds data to an existing Pipe. E.g. DTMF Tone Generator.

     - Data Sink: Terminates a Pipe or extracts data from a Pipe. E.g. DTMF Detector.

     - Data Bridge: Cross-connects data from one Pipe to another or combines data from
     multiple Pipes into one Pipe. E.g. cross-connect in TSS.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_UINT32 nIrpQueueLimit;
  OTX_UINT32 nFifoSize;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create a notification Event and Event Queue for the Raw Data Sender
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataSender[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataSender[i].hNotificationEvent, &g_hLogicDevs.mRawDataSender[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);

    // Create Raw Data Sender
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_DATA_RAW_SENDER, g_hLogicDevs.mRawDataSender[i].hEventQueue, i, &g_hLogicDevs.mRawDataSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    g_hLogicDevs.mRawDataSender[i].nSourceId = i;   // Store sourceId so that we can easily print it from the Sender Thread

    // Increase the IRP Queue limit for fast pipes (to be able to buffer more transmit data in the driver)
    result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, &nIrpQueueLimit); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
    DemoLogf2("RawDataSender[%d]: nIrpQueueLimit = %d\n", i, nIrpQueueLimit);
    
#if 0
    nIrpQueueLimit = 30;
    DemoLogf2("RawDataSender[%d]: Setting nIrpQueueLimit = %d\n", i, nIrpQueueLimit);
    result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, nIrpQueueLimit); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
#endif

#if 0
    // Prevent the driver from generating the OTX_RAWDATA_SENDER_EC_BUF_DEPLETED event 
    // (to reduce driver overhead)
    result = OtxDrvSetAttributeValueBool(
      g_hLogicDevs.mRawDataSender[i].hHandle, 
      OTX_ATTR_RAWDATA_SENDER_GENERATE_DATA_EVENTS, 
      OTX_FALSE); 
    DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
#endif

    // Create a notification Event and Event Queue for the Raw Data Receiver
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent, &g_hLogicDevs.mRawDataReceiver[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, result);

    // Create Raw Data Receiver
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_DATA_RAW_RECEIVER, g_hLogicDevs.mRawDataReceiver[i].hEventQueue, i, &g_hLogicDevs.mRawDataReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    g_hLogicDevs.mRawDataReceiver[i].nSourceId = i; // Store sourceId so that we can easily print it from the Receiver Thread

    // Modify the FIFO size of the Raw Data Receiver
    result = OtxDrvGetAttributeValueUint32(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_ATTR_RAWDATA_RECEIVER_FIFO_SIZE, 
      &nFifoSize); 
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
    DemoLogf2("RawDataReceiver[%d]: nFifoSize = %d bytes  (sizePerTimeslot=%d)\n", i, nFifoSize * (g_demoEnv.nRawDataCapacity/64), nFifoSize);
    if (g_demoEnv.nRawDataCapacity > 256) {
      nFifoSize = 2048;
      DemoLogf2("RawDataReceiver[%d]: Setting nFifoSize = %d bytes  (sizePerTimeslot=%d)\n", i, nFifoSize * (g_demoEnv.nRawDataCapacity/64), nFifoSize);
      result = OtxDrvSetAttributeValueUint32(
        g_hLogicDevs.mRawDataReceiver[i].hHandle, 
        OTX_ATTR_RAWDATA_RECEIVER_FIFO_SIZE, 
        nFifoSize); 
      DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
    }

    // Prevent the driver from generating the OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE event 
    // (to reduce driver overhead)
    result = OtxDrvSetAttributeValueBool(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_ATTR_RAWDATA_RECEIVER_GENERATE_DATA_EVENTS, 
      OTX_FALSE); 
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);

  }

}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];

  // Create the Pipes (one for each direction) between the two DSPs and the two LIs
  for (i = 0; i < g_demoEnv.coLi; i++) {
    OTX_INT32 nCapacity = 64;
    OTX_INT32 nTs;
    
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hLiDspRawData); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hDspLiRawData); DemoCheckRc(g_hPipes[i].hDspLiRawData, result);

    // We will enable the DSP->Li pipe manually when we are ready to send the proper data in the pipe
    result = OtxDrvSetAttributeValueBool(
      g_hPipes[i].hDspLiRawData, 
      OTX_ATTR_PIPE_AUTO_ENABLE, 
      OTX_FALSE
    ); 

    // Connect pipe: Dsp -> Li 
    if (i == 0) {
      nTs = g_demoEnv.nLi0TxTs;
    } else {
      nTs = g_demoEnv.nLi1TxTs;
    }
    DemoLogf1("Li%d: Transmit timeslot = TS%d\n", i, nTs);
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF; 
    result = OtxDrvConnectPipe(g_hPipes[i].hDspLiRawData, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hDspLiRawData, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hDspLiRawData, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hDspLiRawData, result);


    // Connect pipe: Li  -> Dsp 
    if (i == 0) {
      nTs = g_demoEnv.nLi0RxTs;
    } else {
      nTs = g_demoEnv.nLi1RxTs;
    }
    DemoLogf1("Li%d: Receive timeslot  = TS%d\n", i, nTs);
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nTs] = 0xFF; 
    result = OtxDrvConnectPipe(g_hPipes[i].hLiDspRawData, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hLiDspRawData, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
    
  }



  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  // --------------------------------------
 for (i = 0; i < g_demoEnv.coDsp; i++) {
    // Receive direction:
    // Connect Raw Data Receiver
    result = OtxDrvConnectLogicalDevice(g_hPipes[i].hLiDspRawData,  -1,  g_hLogicDevs.mRawDataReceiver[i].hHandle, -1); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
    // Transmit direction:
    // Connect Raw Data Sender
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.mRawDataSender[i].hHandle, -1, g_hPipes[i].hDspLiRawData, -1); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);

  }

 
}


void DemoInitializeLogicalDevices(
/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after 
    the devices are created and connected.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coDsp; i++) {

    DemoLogf2("Initialize mRawDataReceiver[%d] and mRawDataSender[%d]\n", i, i);
    
    // Initialize the Raw Data Receivers (hosted in the DSP)
    result = OtxRawDataReceiverInit(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. Deliver the data as it is received on the highway
      0xFF,        // Flag to be filtered
      OTX_FALSE,   // If OTX_TRUE, filter the specified flag
      OTX_TRUE     // Least Significant part is received first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);

    // Initialize the Raw Data Senders (hosted in the DSP)
    result = OtxRawDataSenderInit(
      g_hLogicDevs.mRawDataSender[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. 
      g_demoEnv.nIdleCode,    // Flag to be sent when the sender is out of data or disabled
      OTX_TRUE     // Least Significant part is transmitted first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
  }

}


void DemoEnableLogicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coDsp; i++) {

    DemoLogf2("Enable and Disable mRawDataSender[%d] (to start sending fillers)\n", i);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
    result = OtxDrvDisable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);

    // We enable the Receiver in the DemoRawDataReceiver thread instead.

  }

}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000;

}



static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


void DemoExit()
{
  g_bExitFlag = OTX_TRUE;
  OsaConsoleExit();
  OtxDrvReleaseLib();
  exit(1);
}

/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf2("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf2("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    if (g_bExitFlag) return 0;
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}


/*******************************************************************************
  @func
    Print buffer usage statistics.
*******************************************************************************/
void *DemoStatsThread(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_UINT32 nCurrentUsage;
  OTX_UINT32 nMaxUsage;
  OTX_UINT32 nLostEvents;
  OTX_UINT32 nQueuedCurrent;
  OTX_UINT32 nQueuedPeak;
  OTX_INT32 nDsp;

  while (g_bStatsThreadActive) {
    
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      
      //result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      //DemoLogf2("  Size (in words)         : %d\n", nAttribValue);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nCurrentUsage); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      //DemoLogf2("  Current usage (in words): %d\n", nAttribValue);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nMaxUsage); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      //DemoLogf2("  Max usage (in words)    : %d\n", nAttribValue);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nLostEvents); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_EXT_QUEUE_CURRENT_USAGE, &nQueuedCurrent); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_EXT_QUEUE_PEAK_USAGE, &nQueuedPeak); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      
      DemoLogf2("Dsp[%d] curr=%.8d, max=%.8d, lost=%d, qcurr=%d, qpeak=%d\n", nDsp, nCurrentUsage, nMaxUsage, nLostEvents, nQueuedCurrent, nQueuedPeak);
      
    }

    OsaThreadSleep(2000); 
  }

  DemoLogf2("DemoStatsThread() terminating.\n");

  return 0;

    
}




/*******************************************************************************
  @func
    Raw Data Receiver Thread Function 
*******************************************************************************/
void *DemoRawDataReceiver(
  void *lpParameter   // thread data
)
{
#define DEMO_RX_BUF_SIZE 8192
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  OTX_UINT8  buf[DEMO_RX_BUF_SIZE];
  OTX_UINT32 nFrameLen;
  OTX_BOOL bFrameEnd;
  //OTX_BOOL bExtractingData = OTX_FALSE; // Only used for flag filtering
  char msgStr[256];
#if !defined(_WIN32_WCE)
  clock_t startTime, stopTime;
#endif  // #if !defined(_WIN32_WCE)
  OTX_UINT32 nSourceId;
  OTX_UINT32 nWaitTime;

#if !defined(_WIN32_WCE)
  startTime = clock();
#endif  // #if !defined(_WIN32_WCE)

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId >= 2) {
    DemoLogf2("DemoRawDataReceiver: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto FATALEND;
  }

  g_hLogicDevs.mRawDataReceiver[nSourceId].bActive  = OTX_TRUE;
  g_hLogicDevs.mRawDataReceiver[nSourceId].bVerbose = OTX_FALSE;

  // Calculate (sleeping) time between each OtxRawDataReadData() call. (256ms fills 2048 bytes in a 64kbit/s pipe)
  //nWaitTime = (256*64)/g_demoEnv.nRawDataCapacity; 
  nWaitTime = (256*64)/g_demoEnv.nRawDataCapacity/4; 

#if 0 // the device will be sync enabled at the same time the sender device is
  // Enable the receiver
  result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);

  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  // Discard any data from the Raw Data Receiver
  do {
    result = OtxRawDataReadData(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, buf, DEMO_RX_BUF_SIZE, &nFrameLen, &bFrameEnd);
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  if (OTX_RESULT_CODE(result) != OTX_S_FALSE) {
    DemoLogf2("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", nSourceId);
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);
  }
#endif

  DemoLogf2("DemoRawDataReceiver[%d]: Waiting for events. Reading data every %d milliseconds.\n", nSourceId, nWaitTime);

  g_ThreadCount++;

WAIT_FOR_EVENTS:

  while (g_hLogicDevs.mRawDataReceiver[nSourceId].bActive) {

    // Wait for a specified time for any (unexpected) events from the Raw Data Receiver device
    OtxDrvWaitForSingleEvent(
      g_hLogicDevs.mRawDataReceiver[nSourceId].hNotificationEvent, 
      nWaitTime
    );
    //DemoLogf2("RawDataReceiver[%d] Received Event  **************\n", nSourceId);
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_RECEIVER:
          // Event from Raw Data Receiver 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf2("RawDataReceiver[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } 

          switch (eventData.m_nCode) {
          case OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE:
            DemoLogf2("RawDataReceiver[%d] Unexpected. Received OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE event.\n", eventData.m_nSourceId);
      			break;
          case OTX_RAWDATA_RECEIVER_EC_SYNC_ENABLE_TRIGGERED:
            DemoLogf1("RawDataReceiver[%d]: Sync Reception started\n", eventData.m_nSourceId);
            break;
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf2("RawDataReceiver[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf2("RawDataReceiver[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
		      }

          break;

        default:
          // Unhandled event
          DemoLogf2("DemoRawDataReceiver[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
          break;
        }

      }

    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

    // Read data from the device
    otxRc = OtxRawDataReadData(
      g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, 
      buf, 
      DEMO_RX_BUF_SIZE, 
      &nFrameLen, 
      &bFrameEnd);
    while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      OTX_INT32 x;
      
      DemoLogf3("RawDataReceiver[%d] Received data (len=%.4d)\n", nSourceId, nFrameLen);

      if (g_hLogicDevs.mRawDataReceiver[nSourceId].bVerbose) {
        for (x = 0; (x < 16) && (x < (OTX_INT32)nFrameLen); x++) {
          DemoLogf2("%.2X ", buf[x]);
        }
        DemoLogf2("\n");
      }
      
      if (g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile) {
        int cbWritten;
        
        cbWritten = fwrite(buf, sizeof(OTX_UINT8), nFrameLen, g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile);
        
        if (cbWritten != (int)nFrameLen) {
          DemoLogf2("RawDataReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", eventData.m_nSourceId, nFrameLen, cbWritten);
        }
        
        if (bFrameEnd) {
          DemoLogf2("RawDataReceiver[%d] Data reception complete. Closing data file. \n", nSourceId);
          fclose(g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile);
          g_hLogicDevs.mRawDataReceiver[nSourceId].pRxFile = NULL;
          goto END;
        }
      }

      if (nFrameLen != DEMO_RX_BUF_SIZE) {
        // No more data to read at this time. Break this while loop
        goto WAIT_FOR_EVENTS;
      }
      
      DemoLogf2("RawDataReceiver[%d] Reading more data. [Abnormal situation]\n", nSourceId);
      otxRc = OtxRawDataReadData(
        g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, 
        buf, 
        DEMO_RX_BUF_SIZE, 
        &nFrameLen, 
        &bFrameEnd);
    }
    if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
      DemoLogf2("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", nSourceId);
      DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, otxRc);
    }
    
  }

END:

#if !defined(_WIN32_WCE)
  stopTime = clock();
#endif  // #if !defined(_WIN32_WCE)

  // Disable the receiver
  result = OtxDrvDisable(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);

#if !defined(_WIN32_WCE)
  DemoLogf2("DemoRawDataReceiver[%d]: Done (time=%2.1f seconds).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
#endif  // #if !defined(_WIN32_WCE)

FATALEND:

  g_ThreadCount--;
  return 0;
}


/*******************************************************************************
  @func
    Raw Data Sender Thread Function 
*******************************************************************************/
void *DemoRawDataSender(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  char msgStr[256];
#if !defined(_WIN32_WCE)
  clock_t startTime, stopTime;
#endif  // #if !defined(_WIN32_WCE)
  OTX_UINT32 nSourceId;
  OTX_UINT32 nTaskRef = 0;
  OTX_INT32 cbPrimed = 0;
  

#if !defined(_WIN32_WCE)
  startTime = clock();
#endif  // #if !defined(_WIN32_WCE)

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId >= 2) {
    DemoLogf("DemoRawDataSender: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto END;
  }

  g_hLogicDevs.mRawDataSender[nSourceId].bActive = OTX_TRUE;
  g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors = 0;

  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  if (g_hLogicDevs.mRawDataSender[nSourceId].pTxFile) {
    // Send from a file
    do {
      g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = fread(g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
      
      otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
      if (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY) {
        DemoLogf3("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
        nTaskRef++;
      }
      
    } while (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf && (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY));
    
  } else {
    // Send a text string (continously until aborted)
    OTX_INT32 cbText = strlen(DEMO_TEST_STRING);
    OTX_INT32 j;

    DemoLogf2("RawDataSender[%d]: Preparing text string buffer...\n", nSourceId);
    
    g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf[0] = '\0';

    // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf). 
    for (j = 0; j < (DEMO_RAW_TX_BUF_SIZE-1)/cbText; j++) {
      strcpy((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf + strlen((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf), DEMO_TEST_STRING);
    }

    g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = strlen((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf);

    DemoLogf2("RawDataSender[%d]: Done preparing text string buffer (len=%d)...\n", nSourceId, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf);

    // Prime the sender
    cbPrimed = 0;
    do {
      otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
      if (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY) {
        cbPrimed += g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf;
        //DemoLogf2("RawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
        nTaskRef++;
      }

    } while (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY);
    DemoLogf2("RawDataSender[%d]: Primed %d bytes\n", nSourceId, cbPrimed);
     
  }
  
  
  while (g_hLogicDevs.mRawDataSender[nSourceId].bActive) {
    OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataSender[nSourceId].hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);
      
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_SENDER:
          // Event from Raw Data Sender 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf("RawDataSender[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } 
          
          switch (eventData.m_nCode) {
          case OTX_RAWDATA_SENDER_EC_COMPLETED:
            if (!g_hLogicDevs.mRawDataSender[nSourceId].pTxFile) {
              g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors++;
              //DemoLogf2("RawDataSender[%d]: *** Data Underrun at cbTxSent/cbTxBuf=%d/%d (m_nParam=%d). \n", nCin, g_sqCin[nCin].cbTxSent, g_sqCin[nCin].cbTxBuf, eventData.m_nParam);
              // Fall-through
            } else {
              DemoLogf2("RawDataSender[%d]: Packet #%d sent\n", eventData.m_nSourceId, eventData.m_nRequestId);
              DemoLogf1("RawDataSender[%d]: Transmission completed.\n", eventData.m_nSourceId);
  
              fclose(g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
              g_hLogicDevs.mRawDataSender[nSourceId].pTxFile = NULL;
              g_hLogicDevs.mRawDataSender[nSourceId].bActive = OTX_FALSE;
              goto END;
            }
            // Fall-through
          case OTX_RAWDATA_SENDER_EC_BUF_DEPLETED:
            {
              OTX_RESULT otxRc;
              
              if (g_hLogicDevs.mRawDataSender[nSourceId].pTxFile) {
                // Sending from file
                DemoLogf3("DemoRawDataSender[%d]: Packet #%d in progress\n", eventData.m_nSourceId, eventData.m_nRequestId);

                if (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                  
                  while (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                    otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
                    if (OTX_RESULT_CODE(otxRc) == OTX_E_BUSY) {
                      //OsaThreadSleep((1024*64)/g_demoEnv.nRawDataCapacity); // Give the driver some time to send the data (1024ms sends 8192 bytes in a 64kbit/s pipe)
                      OsaThreadSleep((1024*64)/g_demoEnv.nRawDataCapacity/2); // Give the driver some time to send the data (1024ms sends 8192 bytes in a 64kbit/s pipe)
                      break;
                    }
                    
                    DemoLogf3("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
                    nTaskRef++;
                    
                    g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = fread(g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
                  }
                  
                  if (!g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                    DemoLogf2("RawDataSender[%d]: EOF reached. Waiting for completion.\n", eventData.m_nSourceId);
                  }
                } else {
                  DemoLogf3("RawDataSender[%d]: Waiting for completion.\n", eventData.m_nSourceId);
                }
              } else if (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
                // Sending from static buffer (continously until aborted)
                while (g_hLogicDevs.mRawDataSender[nSourceId].bActive) {
                  otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
                  if (OTX_RESULT_CODE(otxRc) == OTX_E_BUSY) {
                    OsaThreadSleep((1024*64)/g_demoEnv.nRawDataCapacity); // Give the driver some time to send the data (1024ms sends 8192 bytes in a 64kbit/s pipe)
                    break;
                  }
                  
                  //DemoLogf2("RawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
                  nTaskRef++;
                }
              } else {
                // No file handles and no more bytes to send. Waiting for the OTX_RAWDATA_SENDER_EC_COMPLETED event
              }
            }
            break;
            
          case OTX_RAWDATA_SENDER_EC_SYNC_ENABLE_TRIGGERED:
            DemoLogf1("RawDataSender[%d]: Sync Transmission started\n", eventData.m_nSourceId);
            break;
  
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf2("RawDataSender[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf2("RawDataSender[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
          }
          break;
          default:
            // Unhandled event
            DemoLogf("DemoRawDataSender[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
            break;
        }
        
      }
      OsaThreadSleep(0); // Yield to other threads
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

END:
  
#if !defined(_WIN32_WCE)
  stopTime = clock();
  
  DemoLogf2("RawDataSender[%d]: Done (time=%2.1f seconds, coUnderRunErrors=%d).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors);
#endif  // #if !defined(_WIN32_WCE)
  
  return 0;

}


void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  //EnterCriticalSection(&g_analyzeCritSec);


  DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    
  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf2("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf2("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf2("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf2("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf2("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


  //LeaveCriticalSection(&g_analyzeCritSec);

}


void DemoTransmitAndReceive()
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{

  // Start Receivers
  {
    OTX_INT32  i;
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      char *pszFileName;
      
      if (i == 0) {
        pszFileName = g_demoEnv.pszLi0RxFileName;
      } else {
        pszFileName = g_demoEnv.pszLi1RxFileName;
      }

      if ((g_hLogicDevs.mRawDataReceiver[i].pRxFile = fopen(pszFileName, "wb")) == NULL) {
        DemoLogf2("Error: Could not open receive data file %s \n", pszFileName);
        DemoExit();
      }
      
      DemoLogf1("Li%d: Receiving data file \"%s\"\n", i, pszFileName);
      OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[i].nSourceId);
    }
    
  }

  // Start Senders
  {
    OTX_INT32  i;
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      char *pszFileName;
      
      if (i == 0) {
        pszFileName = g_demoEnv.pszLi0TxFileName;
      } else {
        pszFileName = g_demoEnv.pszLi1TxFileName;
      }
      
      if ((g_hLogicDevs.mRawDataSender[i].pTxFile = fopen(pszFileName, "rb")) == NULL) {
        DemoLogf("Error: Could not open transmit data file %s \n", pszFileName);
        DemoExit();
      }
      
      DemoLogf1("Li%d: Transmit data file  \"%s\"\n", i, pszFileName);
      OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[i].nSourceId);
      
    }

    // Allow the Raw Data Sender threads time to prime the buffers
    OsaThreadSleep(300);
  }


  // Wait for the E1/T1 span to settle
  DemoLogf1("Waiting for the E1/T1 span to settle (%d ms)\n", g_demoEnv.nSetupTime);
  OsaThreadSleep(g_demoEnv.nSetupTime);


  // Enable both senders and receivers synchronously
  {
    OTX_UINT32 nFrameBufferCounter[2];
    OTX_UINT16 nEnablingTriggerPoint;
    OTX_INT32  i;
    OTX_RESULT result;
    
    i = 0;
    result = OtxDrvGetAttributeValueUint32(
      g_hPhysDevs.hDsp[i], 
      OTX_ATTR_DSP_FRAME_BUFFER_COUNTER, 
      &nFrameBufferCounter[i]
      ); 
    i = 1;
    result = OtxDrvGetAttributeValueUint32(
      g_hPhysDevs.hDsp[i], 
      OTX_ATTR_DSP_FRAME_BUFFER_COUNTER, 
      &nFrameBufferCounter[i]
      ); 
    DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    
    DemoLogf2("DSP[%d]: FrameBufferCounter    = %.6d (0x%.4X)\n", 0, nFrameBufferCounter[0], nFrameBufferCounter[0]);
    DemoLogf2("DSP[%d]: FrameBufferCounter    = %.6d (0x%.4X)\n", 1, nFrameBufferCounter[1], nFrameBufferCounter[1]);
    
    // In the OTXHDLC SPM the nFrameBufferCounter is incremented every 2.5ms (20 frames).
    // Trigger the enabling about 25 ms from now (i.e. nFrameBufferCounter + 30).
    nEnablingTriggerPoint = (OTX_UINT16)(nFrameBufferCounter[0] + 200 /*40*/ /*10*/);
    if (nEnablingTriggerPoint >= 0xFFFE) {
      // Invalid value. Step one more (wrap)
      nEnablingTriggerPoint = 0;
    }
    
    DemoLogf2("DSP[%d]: nEnablingTriggerPoint = %.6d (0x%.4X)\n", i, nEnablingTriggerPoint, nEnablingTriggerPoint);
    
    // Set the enabling trigger point
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvSetAttributeValueUint32(
        g_hLogicDevs.mRawDataReceiver[i].hHandle, 
        OTX_ATTR_DEVICE_SYNC_ENABLE, 
        nEnablingTriggerPoint+1   // +1 since it will take the sender 20 frames to actually start the transmisson (due to the 20 frame buffering in the DSP)
        ); 
      DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      result = OtxDrvSetAttributeValueUint32(
        g_hLogicDevs.mRawDataSender[i].hHandle, 
        OTX_ATTR_DEVICE_SYNC_ENABLE, 
        nEnablingTriggerPoint
        ); 
      DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    
    // Enable the devices (they will enable when they reach the trigger point)
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      DemoLogf2("Enable mRawDataSender[%d] and mRawDataReceiver[%d]\n", i, i);
      result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
#if 1 // Just to simulate a slow PC and to demonstrate that the devices are still synchronized despite this delay.
      OsaThreadSleep(1);
#endif
      result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
#if 1 // Just to simulate a slow PC and to demonstrate that the devices are still synchronized despite this delay.
      OsaThreadSleep(1);
#endif
    }
  }

  // Wait for both senders to complete
  {
    OTX_INT32  i;

    for (;;) { // Loop exited by return statement or by Ctrl-Break
      
      if (g_hLogicDevs.mRawDataSender[0].bActive == OTX_FALSE &&
        g_hLogicDevs.mRawDataSender[1].bActive == OTX_FALSE
        ) {

        // Wait for the receiver to pickup the last data
        OsaThreadSleep(g_demoEnv.nHoldTime);

        for (i = 0; i < g_demoEnv.coDsp; i++) {
          DemoLogf2("Aborting Raw Data Receiver[%d] Thread\n", i);
          g_hLogicDevs.mRawDataReceiver[i].bActive = OTX_FALSE;
        }

        // try to exit cleanly, wait a few seconds for threads to complete
        for (i = 0; i < 5; i++) {
          DemoLogf1("Waiting for %d threads to exit.\n", g_ThreadCount);
          if (g_ThreadCount == 0) {
            break;
          }
          OsaThreadSleep(500);
        }
        if (g_ThreadCount) {
          DemoLogf1("Error!  %d threads still active\n", g_ThreadCount);
        }

        return;
      }
      OsaThreadSleep(200);
    }
  }


}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  DemoLogf1("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf1("Cmd  Synopsis\n");
    DemoLogf1("===  ========\n");
    DemoLogf1(" ?   Display Help text\n");
    DemoLogf1(" h   Ditto\n");
    DemoLogf1(" q   Quit\n");
    DemoLogf1(" 0-1 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    DemoLogf1(" x   Send files on both Lis and Receive on both Lis\n");
    DemoLogf1(" a   Abort all Raw Data Sender and Receiver threads\n");
    DemoLogf1(" r   Start Raw Data Receiver[%d] (no file logging)\n", nCurrentLi);
    DemoLogf1(" R   Start Raw Data Receiver[%d]. Prompt for file name\n", nCurrentLi);
    DemoLogf1(" s   Start Raw Data Sender[%d]. Continously send a text string\n", nCurrentLi);
    DemoLogf1(" S   Start Raw Data Sender[%d]. Prompt for file to send\n", nCurrentLi);
    DemoLogf1(" v   Toggle verbose (print received data) setting Raw Data Receiver[%d]\n", nCurrentLi);
    DemoLogf1(" l   Query Line Status of both Lis\n");
    DemoLogf1(" t   Toggle verbose buffer usage statistics\n");
    DemoLogf1(" e   Synchronized enabling of the Raw Data Senders\n");
    DemoLogf1(" d/D Dump debug information to kernel Debugger for DSP0/1\n");
    break;


  case 'x':
    {
      DemoTransmitAndReceive();
    }
    break;

  case 'g':
    {
      OTX_INT32  i;
      OTX_INT32  j;

      DemoLogf1("Sync Test (View with logical analyzer)\n");
      
      for (j = 0; j < 200; j++) {
        //DemoLogf1(".");
        DemoLogf1("j=%d ", j);
        
        for (i = 0; i < g_demoEnv.coDsp; i++) {
          // Initialize the Raw Data Senders (hosted in the DSP)
          result = OtxRawDataSenderInit(
            g_hLogicDevs.mRawDataSender[i].hHandle, 
            OTX_DATA_FORMAT_UINT8,  // No data conversion. 
            (OTX_UINT8) (((j%2) == 0) ? 0xF0 : 0x0F),        // Flag to be sent when the sender is out of data or disabled
            OTX_TRUE     // Least Significant part is transmitted first. Only applicable for a subchannel pipe (nCapacity < 64).
            ); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
        }
        
        {
          OTX_UINT32 nFrameBufferCounter;
          OTX_UINT16 nEnablingTriggerPoint;
          OTX_INT32  i;
          
          i = 0;
          result = OtxDrvGetAttributeValueUint32(
            g_hPhysDevs.hDsp[i], 
            OTX_ATTR_DSP_FRAME_BUFFER_COUNTER, 
            &nFrameBufferCounter
            ); 
          DemoCheckRc(g_hPhysDevs.hDsp[i], result);
          
          //DemoLogf1("DSP[%d]: FrameBufferCounter    = %.6d (0x%.4X)\n", i, nFrameBufferCounter, nFrameBufferCounter);
          
          // In the OTXHDLC SPM the nFrameBufferCounter is incremented every 2.5ms (20 frames).
          // Trigger the enabling about 25*3=75 ms from now (i.e. nFrameBufferCounter + 30).
          nEnablingTriggerPoint = (OTX_UINT16)(nFrameBufferCounter + 30);
          if (nEnablingTriggerPoint == 0xFFFF) {
            // Invalid value. Step one more (wrap)
            nEnablingTriggerPoint = 0;
          }
          
          DemoLogf1(" 0x%.4X) ", nEnablingTriggerPoint);
          
          // Set the enabling trigger point
          for (i = 0; i < g_demoEnv.coDsp; i++) {
            result = OtxDrvSetAttributeValueUint32(
              g_hLogicDevs.mRawDataSender[i].hHandle, 
              OTX_ATTR_DEVICE_SYNC_ENABLE, 
              nEnablingTriggerPoint
              ); 
            DemoCheckRc(g_hPhysDevs.hDsp[i], result);
          }
          
          // Enable the devices (they will enable when they reach the trigger point)
          for (i = 0; i < g_demoEnv.coDsp; i++) {
            result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
          }
          
          for (i = 0; i < g_demoEnv.coDsp; i++) {
            OtxEventDataS eventData;
            // Wait until the Enable command is "triggered"
            OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataSender[i].hNotificationEvent, 4000);
            result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[i].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);
            if (eventData.m_nCode == OTX_RAWDATA_SENDER_EC_SYNC_ENABLE_TRIGGERED) {
              DemoLogf1(":");
            } else {
              DemoLogf1("Error: Unexpected event from Raw Data Sender[%d]", i);
            }
          }
          
          for (i = 0; i < g_demoEnv.coDsp; i++) {
            result = OtxDrvDisable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
          }
          
          DemoLogf1(";");

          OsaThreadSleep(200);

          DemoLogf1(";\n");

        }
        
      }
      
      DemoLogf1("\nDone\n");
    }
    break;

  case 'd':
    {
      OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 0);
    }
    break;

  case 'D':
    {
      OtxDrvDbgDump(g_hPhysDevs.hDsp[1], 0);
    }
    break;

  case 'e':
    {
      OTX_UINT32 nFrameBufferCounter;
      OTX_UINT16 nEnablingTriggerPoint;
      OTX_INT32  i;
      
      i = 0;
      result = OtxDrvGetAttributeValueUint32(
        g_hPhysDevs.hDsp[i], 
        OTX_ATTR_DSP_FRAME_BUFFER_COUNTER, 
        &nFrameBufferCounter
        ); 
      DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      
      DemoLogf1("DSP[%d]: FrameBufferCounter    = %.6d (0x%.4X)\n", i, nFrameBufferCounter, nFrameBufferCounter);
      
      // In the OTXHDLC SPM the nFrameBufferCounter is incremented every 2.5ms (20 frames).
      // Trigger the enabling about 25 ms from now (i.e. nFrameBufferCounter + 10).
      nEnablingTriggerPoint = (OTX_UINT16)(nFrameBufferCounter + 10);
      if (nEnablingTriggerPoint == 0xFFFF) {
        // Invalid value. Step one more (wrap)
        nEnablingTriggerPoint = 0;
      }

      DemoLogf1("DSP[%d]: nEnablingTriggerPoint = %.6d (0x%.4X)\n", i, nEnablingTriggerPoint, nEnablingTriggerPoint);

      // Set the enabling trigger point
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        result = OtxDrvSetAttributeValueUint32(
          g_hLogicDevs.mRawDataSender[i].hHandle, 
          OTX_ATTR_DEVICE_SYNC_ENABLE, 
          nEnablingTriggerPoint
          ); 
        DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      }

      // Enable the devices (they will enable when they reach the trigger point)
      i = 0;
      DemoLogf1("Enable mRawDataSender[%d]\n", i);
      result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
      DemoLogf1("Enable mRawDataReceiver[%d]\n", i);
      result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
#if 1      
      i = 1;
      DemoLogf1("Enable mRawDataSender[%d]\n", i);
      result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
      DemoLogf1("Enable mRawDataReceiver[%d]\n", i);
      result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
#endif      
    }
    break;


  case 't':
    {
      if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
        DemoLogf1("Function not supported by this board type\n");
      } else {
        g_bStatsThreadActive = !g_bStatsThreadActive;
        
        if (g_bStatsThreadActive) {
          DemoLogf1("Starting Stats Thread\n");
          OsaThreadStart(DemoStatsThread);
        }
      }
    }
    break;


  case 'a':
    {
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coDsp; i++) {
        

        DemoLogf1("Aborting Raw Data Receiver[%d] Thread\n", i);
        g_hLogicDevs.mRawDataReceiver[i].bActive = OTX_FALSE;

        
        DemoLogf1("Aborting Raw Data Sender[%d] Thread\n", i);
        g_hLogicDevs.mRawDataSender[i].bActive = OTX_FALSE;
      }

    }
    break;

  case 'v':
    {

      g_hLogicDevs.mRawDataReceiver[nCurrentLi].bVerbose = !g_hLogicDevs.mRawDataReceiver[nCurrentLi].bVerbose;
      DemoLogf1("Setting verbose setting to %s for Raw Data Receiver[%d] Thread\n", (g_hLogicDevs.mRawDataReceiver[nCurrentLi].bVerbose)?"ON":"OFF", nCurrentLi);

    }
    break;


  case 'r':
    {


      g_hLogicDevs.mRawDataReceiver[nCurrentLi].pRxFile = NULL;

      DemoLogf1("Creating Raw Data Receiver[%d] Thread\n", nCurrentLi);
    	OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[nCurrentLi].nSourceId);

    }
    break;

  case 'R':
    {
      //OTX_RESULT otxRc = OTX_S_OK;
      char sqzFileName[64];
      OtxDeviceStateE  eDevState;

      DemoLogf1("RawDataReceiver[%d]. Please enter file name: ", nCurrentLi);
      OsaConsoleGetStr(sqzFileName);


      if ((g_hLogicDevs.mRawDataReceiver[nCurrentLi].pRxFile = fopen(sqzFileName, "wb")) == NULL) {
        DemoLogf1("Error: Could not open file %s \n", sqzFileName);
        break;
      }

      DemoLogf1("Creating Raw Data Receiver[%d] Thread\n", nCurrentLi);
    	OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[nCurrentLi].nSourceId);

      OsaThreadSleep(200);
      OtxDrvGetState(g_hLogicDevs.mRawDataReceiver[nCurrentLi].hHandle, &eDevState);
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf1("To start receiving data, press the 'e' key to enable the device\n");
      }

    }
    break;


  case 'S':
    {
      //OTX_RESULT otxRc = OTX_S_OK;
      char sqzFileName[64];
      OtxDeviceStateE  eDevState;

      DemoLogf1("RawDataSender[%d]. Please enter file to send: ", nCurrentLi);
      OsaConsoleGetStr(sqzFileName);

      if ((g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = fopen(sqzFileName, "rb")) == NULL) {
        DemoLogf1("Error: Could not open file %s \n", sqzFileName);
        break;
      }

      DemoLogf1("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
    	OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
    
      OsaThreadSleep(200);
      OtxDrvGetState(g_hLogicDevs.mRawDataSender[nCurrentLi].hHandle, &eDevState);
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf1("To start sending data, press the 'e' key to enable the device\n");
      }


    }
    break;



  case 's':
    {
      DemoLogf1("RawDataSender[%d]. Continously send text string \"%s\".\n", nCurrentLi, DEMO_TEST_STRING);

      g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = NULL;

      DemoLogf1("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
    	OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
    
    }
    break;



  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf1("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf1("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;



  case '0':
    // Select Li/Dsp 0 as current
    nCurrentLi = 0;
    DemoLogf1("Li/Dsp#0 Selected ....\n");
    break;

  case '1':
    // Select Li/Dsp 1 as current
    nCurrentLi = 1;
    DemoLogf1("Li/Dsp#1 Selected ....\n");
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf1("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}                      




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;
  
  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_T1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_PCMCIA;
  envP->nRawDataCapacity  = 64;
  envP->eProgramModule    = OTX_DSP_SPM_OTXHDLC;
  envP->coLi              = 2;
  envP->coDsp             = DEMO_DSP_MAX;
  envP->pszLi0TxFileName  = DEMO_DEFAULT_LI0_TX_FILE_NAME;
  envP->pszLi1TxFileName  = DEMO_DEFAULT_LI1_TX_FILE_NAME;
  envP->pszLi0RxFileName  = DEMO_DEFAULT_LI0_RX_FILE_NAME;
  envP->pszLi1RxFileName  = DEMO_DEFAULT_LI1_RX_FILE_NAME;
  envP->nLi0TxTs          = 1;
  envP->nLi1TxTs          = 1;
  envP->nLi0RxTs          = 1;
  envP->nLi1RxTs          = 1;
  envP->bInteractiveMode  = OTX_FALSE;
  envP->nIdleCode         = 255;
  envP->nLogVerbosity     = 1;
  envP->nHoldTime         = DEMO_DEFAULT_HOLD_TIME; // in milliseconds
  envP->nSetupTime        = DEMO_DEFAULT_SETUP_TIME; // in milliseconds
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->bReceiveEqualizer = OTX_FALSE; 
  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->bConfigure        = OTX_FALSE;
  
  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'A') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);

        } else if (*(argv[i]+1) == 'O') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;

        } else if (*(argv[i]+1) == 'V') {   // Log verbosity
          envP->nLogVerbosity = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'W') {   // Configure and exit
          envP->bConfigure = OTX_TRUE;

        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'H') {   // Hold time
          envP->nHoldTime = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'S') {   // Setup time
          envP->nSetupTime = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'I') {   // Idle code
          envP->nIdleCode = (OTX_UINT8)strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'K') {   // Interactive mode
          envP->bInteractiveMode = OTX_TRUE;

        } else if (*(argv[i]+1) == 'Q') {   // Receive Equalizer (Long-haul mode)
          envP->bReceiveEqualizer = OTX_TRUE;

        } else if (*(argv[i]+1) == 'N') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi  = 1;
            envP->coDsp = 1; // This really means that the board has one burst device. We use Dsp in lieu of Burst device.
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'D') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'Z') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'T') {   // Tx slot
          if ( *(argv[i]+2) == 'A') {         // Li0
            envP->nLi0TxTs = strtoul(argv[i]+3, NULL, 0);
          } else if ( *(argv[i]+2) == 'B') {  // Li1
            envP->nLi1TxTs = strtoul(argv[i]+3, NULL, 0);
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'R') {   // Rx slot
          if ( *(argv[i]+2) == 'A') {         // Li0
            envP->nLi0RxTs = strtoul(argv[i]+3, NULL, 0);
          } else if ( *(argv[i]+2) == 'B') {  // Li1
            envP->nLi1RxTs = strtoul(argv[i]+3, NULL, 0);
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'P') {   // Tx file
          if ( *(argv[i]+2) == 'A') {         // Li0
            envP->pszLi0TxFileName = argv[i]+3;
          } else if ( *(argv[i]+2) == 'B') {  // Li1
            envP->pszLi1TxFileName = argv[i]+3;
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'G') {   // Rx file
          if ( *(argv[i]+2) == 'A') {         // Li0
            envP->pszLi0RxFileName = argv[i]+3;
          } else if ( *(argv[i]+2) == 'B') {  // Li1
            envP->pszLi1RxFileName = argv[i]+3;
          } else {
            DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf1("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   // Line Interface 0
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   // Line Interface 1
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '3') {   // Split clock, both Lis synchronized to its own line
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   // Split clock, Li0 synchronized to its T1E1 span. Li1 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   // Split clock, Li1 synchronized to its T1E1 span. Li0 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else {
          DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }

      } else {
        DemoLogf1("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
  
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>  : Log file name (default is NUL)\n");
  DemoLogf(" /B<boardNo>   : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /N[2|T|E|4|8|M|P|S]: Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI with Vidar-5x4-ASM \n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus with Vidar-5x4 Plus \n");
  DemoLogf("                  T = Use Thor-2-PCI-Express with Vidar-5x4 Plus \n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus with Vidar-5x4 Plus \n");
  DemoLogf("                  8 = Use Thor-8-PCI with Vidar-5x4-ASM \n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus with Vidar-5x4 Plus \n");
  DemoLogf("                  S = Use Sleipnir-1-PCI-Plus \n");
  DemoLogf(" /W            : Configure Li and exit\n");
  DemoLogf(" /ME           : E1 Li mode\n");
  DemoLogf(" /MT           : T1 Li mode (default)\n");
  DemoLogf(" /K            : Interactive mode\n");
  DemoLogf(" /V<level>     : Log verbosity level (0-5). Default is 1. \n"
           "                 0=No log. 5=Most verbose.\n");
  DemoLogf(" /I<value>     : Idle value to transmit when no data is\n" 
           "                 transmitted (0-255). Default is 255.\n");
  DemoLogf(" /S<time>      : Setup time. The amount of time (in ms) the application will\n"
           "                 sleep before data transmission/reception is started.\n"
           "                 Default is %d milliseconds.\n", DEMO_DEFAULT_SETUP_TIME);
  DemoLogf(" /H<time>      : Hold time. The amount of time (in ms) the application will\n"
           "                 receive data after the last data byte is transmitted.\n"
           "                 Default is %d milliseconds.\n", DEMO_DEFAULT_HOLD_TIME);
  DemoLogf(" /TA<timeslot> : Transmit timeslot for Li0 (default is TS1)\n");
  DemoLogf(" /TB<timeslot> : Transmit timeslot for Li1 (default is TS1)\n");
  DemoLogf(" /RA<timeslot> : Receive timeslot for Li0 (default is TS1)\n");
  DemoLogf(" /RB<timeslot> : Receive timeslot for Li1 (default is TS1)\n");
  DemoLogf(" /PA<filename> : Transmit file Li0 (default is %s)\n", DEMO_DEFAULT_LI0_TX_FILE_NAME);
  DemoLogf(" /PB<filename> : Transmit file Li1 (default is %s)\n", DEMO_DEFAULT_LI1_TX_FILE_NAME);
  DemoLogf(" /GA<filename> : Receive file Li0 (default is %s)\n", DEMO_DEFAULT_LI0_RX_FILE_NAME);
  DemoLogf(" /GB<filename> : Receive file Li1 (default is %s)\n", DEMO_DEFAULT_LI1_RX_FILE_NAME);
  DemoLogf(" /Q            : Long-haul mode (default is Short-haul)\n");
  DemoLogf(" /F[E|S|9]     : T1 Frame Format, where\n");
  DemoLogf("                  E = ESF (default)\n");
  DemoLogf("                  S = SF [D3/D4]\n");
  DemoLogf("                  9 = SLC96\n");
  DemoLogf(" /J[A|B]       : T1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  B = B8ZS (default)\n");
  DemoLogf(" /D[D|C|G]     : E1 Frame Format, where\n");
  DemoLogf("                  D = Doubleframe\n");
  DemoLogf("                  C = CRC4 Multiframe (default)\n");
  DemoLogf("                  G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /Z[A|H]       : E1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  H = HDB3 (default)\n");
  DemoLogf(" /C[I|0|1|3|4|5]:Clock source, where\n");
  DemoLogf("                  I = Internal board oscillator as clock source\n");
  DemoLogf("                  0 = Li0 recovered clock (default)\n");
  DemoLogf("                  1 = Li1 recovered clock\n");
  DemoLogf("                  3 = Split Clock, both Line Interface are\n"
           "                      synchronized to its T1E1 span\n");
  DemoLogf("                  4 = Split Clock, Li0 synchronized to its T1E1 span.\n"
           "                      Li1 internal clock\n");
  DemoLogf("                  5 = Split Clock, Li1 synchronized to its T1E1 span.\n"
           "                      Li0 internal clock\n");
  DemoLogf(" /A<str>       : Command string (overrides keyboard input)\n");
  DemoLogf(" /O           : Exit without prompting on errors\n");
}


#define DEMOLOGF \
{ \
  va_list argList; \
  OsaMutexLock(&g_demoLogfCritSec); \
  va_start(argList, formatStr); \
  OsaConsolePrintF(formatStr, argList); \
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList); \
  va_end(argList); \
  OsaMutexUnlock(&g_demoLogfCritSec); \
}

static void DemoLogf1(char *formatStr, ...)
{
  if (g_demoEnv.nLogVerbosity < 1) return;
  DEMOLOGF;
}

static void DemoLogf2(char *formatStr, ...)
{
  if (g_demoEnv.nLogVerbosity < 2) return;
  DEMOLOGF;
}

static void DemoLogf3(char *formatStr, ...)
{
  if (g_demoEnv.nLogVerbosity < 3) return;
  DEMOLOGF;
}

/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  DEMOLOGF;
}



int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf1("T2mRawDataSync, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf2("Performs synchronized Raw Data sending and receiving for Li0 and Li1.\n");

  DemoLogf2("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf2("Raw Data Pipe capacity is %d\n", g_demoEnv.nRawDataCapacity);

  if ((g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1) && (g_demoEnv.nRawDataCapacity != 64)) {
    DemoLogf2("Error: OTX_DSP_SPM_OTXSPM1 only supports 64kpbs pipes in this demo. Exiting.\n");
    OsaConsoleExit();
    exit(-1);
  }

  
  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf2("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf2("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf2("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf2("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf2("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf2("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  if (g_demoEnv.bConfigure) {
    // Set clock source for the board 
    if (g_hPhysDevs.hVidarAsmController0) {
      // Thor-2-PCI
      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    } else {
      // Thor-2-PCMCIA, Sleipnir-1-PCI
      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
  }

    // Configure E1/T1 devices
  DemoLogf2("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf2("Configured E1/T1 devices\n");

  // configuration mode only, we're finished
  if (g_demoEnv.bConfigure) {
    DemoExit();
  }

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf2("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf2("DemoCreateLogicalDevices\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf2("DemoConnectDevices\n");
  DemoConnectDevices();

  // Initialize the logical devices. A logical device must be create and connected before it can be initialized.
  DemoLogf2("DemoInitializeLogicalDevices\n");
  DemoInitializeLogicalDevices();

  // Enable logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf2("DemoEnableLogicalDevices\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  // Reinitialize the serial ports of each DSP
  {
    OTX_INT32 nDsp;
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
    }
  }

  // Enable DSP Serial port clocks. All DSPs are now running synchronously.
  result = OtxDrvSetAttributeValueBool(
    g_hPhysDevs.hThorController0, 
    OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
    OTX_FALSE
    ); 

  OsaThreadSleep(200); // Need to sleep here to make sure that the DSP serial ports 
                       // pump out any non-idle characters out its serial port 
                       // registers before we enable the pipe

  {
    OTX_INT32 i;
    OTX_RESULT result;
    
    // Enable the DSP -> LI pipe (i.e. allow the DSP to output data on the pipe)
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvEnable(g_hPipes[i].hDspLiRawData, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPipes[i].hDspLiRawData, result);
      
    }
    
  }

  if (g_demoEnv.bInteractiveMode) {

    DemoLogf1("\nPress 'h' for help\n");
    
    while (!exitFlag) {
      commandChar = OsaConsoleGetCh();
      exitFlag = DemoProcessUserCommand(commandChar);
    }
  } else {
    //OsaThreadSleep(1000); // Allow T1 or E1 to stabilize
    DemoTransmitAndReceive();
    DemoLogf1("Exiting...\n");
    DemoExit();
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/ME", "/K", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
